
#ifndef __TCOMLUT3D__
#define __TCOMLUT3D__

#include "TypeDef.h"

#define FIXED_ARRAY     (0)
#define USE_SHIFT_DIV   (1)
#define USE_IBD         (1)

#define INTERPOLATION_TRILINEAR   (0)
#define INTERPOLATION_TETRAHEDRAL (1)

#if FIXED_ARRAY
  #define NBP 5
#endif
#define BOX_MAX 256

typedef struct {
  Short sY, sU, sV;
  Bool  bEncoded;
  Bool  bUsed;
  Int   iLayer;     // 0: Top layer
} node; // node

typedef struct {
  Bool  bSplit;
  Void  *leave[8];
} octTree;  // octant or cube

typedef struct {
  Bool  bFlag;  // 1: flag, 0: SVLC
  Int   iValue;
} syntaxElement;  // octant or cube

class  TComLut3D
{
  Int     m_nbp;
#if USE_SHIFT_DIV
  Int     m_dr;
  Int     m_dg;
  Int     m_db;
  Int     m_shift_div;  // = log(m_nbp)/log(2)
  Int     m_shift_div3; // = 3*m_shift_div

#if USE_IBD
  Int     m_shift_divY, m_shift_divC; // used by interpol_ibd
  Int     m_shift_div3Y, m_shift_div3C;
#endif
#endif
#if FIXED_ARRAY
  node    m_yuvLut[NBP][NBP][NBP];
#else
  node    *m_yuvLut;
#endif
  octTree *m_octTree;

  Void  interpolLut_fromUpperLayer( node *nodeIn, node *nodeOut, Int iUpperLayer );
  Void  encode(Int iLayer, int r0, int g0, int b0, int errMax, Int errQ, TComLut3D *lutRef, TComLut3D *lutResiduals, octTree *pOctTree, octTree *pOctTreeParent, Bool bVerbose );

public:
  TComLut3D( Int iNbpCode, Bool bAllocateOctree=false );
  ~TComLut3D();

  inline  node    *getNode( int r, int g, int b );
  Void    interpol( node *nodeIn, node *nodeOut );
  Void    interpol_tetra( node *nodeIn, node *nodeOut );
  Void    reset( Bool bResetNode=true, Bool bResetFlag=true, Bool bUsedFlag=false );
  Void    setDefault();
#if USE_IBD
  Void    set_ibd( Int internalBitDepthY, Int internalBitDepthC );
  Void    interpol_ibd( node *nodeIn, node *nodeOut );
#endif

  Void    encode( Int errMax, Int errQ, TComLut3D *lutRef, TComLut3D *lutResiduals, Bool bWriteStream );
  Void    decode( Int iLayer, int r0, int g0, int b0, Int errQ, TComLut3D* lutres, octTree *pOctTree, Bool bVerbose );
  Int     getNbpCode();
  Void    prune( Bool b );

  inline Int      getNbp()      { return m_nbp; };
  inline octTree* getOctTree()  { return m_octTree; };
};

inline  node *TComLut3D::getNode( int r, int g, int b )
{
#if FIXED_ARRAY
  return( &m_yuvLut[r][g][b] );
#else
  return( m_yuvLut + r*m_nbp*m_nbp + g*m_nbp + b );
#endif
}

#endif

