/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComWeightPrediction.h
    \brief    weighting prediction class (header)
*/

// Include files
#include "TComToneMapping.h"
#include "TComSlice.h"
#include "TComLut3D.h"

static  UChar nbpCode[7]={ 0, 2, 3, 5, 9, 17, 33 };

// ====================================================================================================================
// Class definition
// ====================================================================================================================
TComToneMapping::TComToneMapping( Int iModel, Int iNbpCode, Char *cToneMappingFile )
{
  m_colour_map_id = 0;
  m_colour_map_cancel_flag = false;
  m_colour_map_repetition_period = 1;
  m_colour_map_video_signal_type_present_flag = true;
  m_colour_map_video_format = 5;
  m_colour_map_video_full_range_flag = false;
  m_colour_map_description_present_flag = true;
  m_colour_map_primaries = 2;
  m_colour_map_transfer_characteristics = 2;
  m_colour_map_matrix_coeffs = 2;
  m_bit_depth_lut_minus8 = 0;

  m_bVerbose        = false;

  m_toneMappingFile = cToneMappingFile;
  m_nbpCode         = iNbpCode;

  m_lutRef        = NULL;
  m_lutCoded      = NULL;
  m_lutResiduals  = NULL;

  if ( cToneMappingFile ) loadFile( cToneMappingFile );
  else                    m_model = iModel;

	m_pcTempPic	= NULL;
  m_errQ      = 1;

  if (m_model == MODEL_3DLUT)
  {
    if ( !cToneMappingFile )
      m_lutRef      = new TComLut3D( getNbp() );
    m_lutCoded      = new TComLut3D( getNbp(), true );
    m_lutResiduals  = new TComLut3D( getNbp() );
  }
  else
  {
    printf("TComToneMapping::TComToneMapping():\tUnknown model.\n");
    exit(0);
  }
}

TComToneMapping::~TComToneMapping()
{
	delete m_pcTempPic;
	if ( m_lutRef )       delete m_lutRef;
	if ( m_lutCoded )     delete m_lutCoded;
	if ( m_lutResiduals ) delete m_lutResiduals;
}

Void  TComToneMapping::writeFile( Char *cFile )
{
  Char  cTmp[500];

  sprintf( cTmp, "%s_3Dlut_nbp%d.txt", cFile, getNbp() );

  FILE  *fic = fopen( cTmp, "w" );

  fprintf( fic, "%s\n", "3DLUT");

  // 3D LUT model
  fprintf(fic, "%d\n", m_colour_map_id);
  fprintf(fic, "%d\n", m_colour_map_cancel_flag ? 1 : 0); 
  if (m_colour_map_cancel_flag == false)
  {
    fprintf(fic, "%d\n", m_colour_map_repetition_period);
    fprintf(fic, "%d\n", m_colour_map_video_signal_type_present_flag ? 1 : 0);
    if (m_colour_map_video_signal_type_present_flag)
    {
      fprintf(fic, "%d\n", m_colour_map_video_format);
      fprintf(fic, "%d\n", m_colour_map_video_full_range_flag ? 1 : 0);
      fprintf(fic, "%d\n", m_colour_map_description_present_flag ? 1 : 0);
      if (m_colour_map_description_present_flag)
      {
        fprintf(fic, "%d\n", m_colour_map_primaries);
        fprintf(fic, "%d\n", m_colour_map_transfer_characteristics);
        fprintf(fic, "%d\n", m_colour_map_matrix_coeffs);
      }
    }

    fprintf(fic, "%d\n", m_bit_depth_lut_minus8);
    fprintf(fic, "%d\n", m_nbpCode);
    for ( Int r=0 ; r<getNbp() ; r++ )
    for ( Int g=0 ; g<getNbp() ; g++ ) {
      for ( Int b=0 ; b<getNbp() ; b++ ) {
        node* n = m_lutRef->getNode( r,g,b );
        fprintf(fic, "%d ",  n->sY);
        fprintf(fic, "%d ",  n->sU);
        fprintf(fic, "%d\t\t", n->sV);
      }
      fprintf(fic, "\n");
    }
  }

  fclose(fic);
}

Void  TComToneMapping::loadFile( Char *cFile )
{
  FILE  *fic = fopen( cFile, "r" );

  Char  format[100];
  fscanf( fic, "%s", format );
  Int flag = 0;

  fscanf(fic, "%d", &m_colour_map_id);
  fscanf(fic, "%d", &flag); 
  m_colour_map_cancel_flag = (flag == 1) ? true : false;
  if (m_colour_map_cancel_flag == false)
  {
    fscanf(fic, "%d", &m_colour_map_repetition_period);
    fscanf(fic, "%d", &flag);
    m_colour_map_video_signal_type_present_flag = (flag == 1) ? true : false;
    if (m_colour_map_description_present_flag)
    {
      fscanf(fic, "%d", &m_colour_map_video_format);
      fscanf(fic, "%d", &flag);
      m_colour_map_video_full_range_flag = (flag == 1) ? true : false;
      fscanf(fic, "%d", &flag);
      m_colour_map_description_present_flag = (flag == 1) ? true : false;
      if (m_colour_map_description_present_flag)
      {
        fscanf(fic, "%d", &m_colour_map_primaries);
        fscanf(fic, "%d", &m_colour_map_transfer_characteristics);
        fscanf(fic, "%d", &m_colour_map_matrix_coeffs);
      }
    }

    // 3D LUT model
    if ( strcmp(format,"3DLUT")==0 ) {
      m_model = MODEL_3DLUT;
      fscanf(fic, "%d", &m_bit_depth_lut_minus8);
      fscanf(fic, "%d", &m_nbpCode);
      m_lutRef  = new TComLut3D( getNbp() );
      for ( Int r=0 ; r<getNbp() ; r++ )
      for ( Int g=0 ; g<getNbp() ; g++ )
      for ( Int b=0 ; b<getNbp() ; b++ ) {
        node* n = m_lutRef->getNode( r,g,b );
        fscanf(fic, "%hi", &n->sY);
        fscanf(fic, "%hi", &n->sU);
        fscanf(fic, "%hi", &n->sV);
      }
    }
    else {
      printf("TComToneMapping::loadFile():\tUnknown model.\n");
      exit(0);
    }
  }

  fclose(fic);
}

Void	TComToneMapping::analysis( TVideoIOYuv& pTVideoIOYuvInputFile0, TComPicYuv* pPicYuv0, Int *aiPad0, TComSPS *sps, Int iPeriod, Int iNbFrames )
{
  Int nbSkippedFrames=0;

  //--------------------------------------------
  //  Pruning
  //--------------------------------------------
  if ( m_model == MODEL_3DLUT ) {
    m_lutRef->reset( false, true );

    Int nbFrameAnalysis = iNbFrames / iPeriod ;
    nbFrameAnalysis     = ( nbFrameAnalysis<=0 ) ? (1) : ( nbFrameAnalysis );
    Int nbSkipFrames    = iPeriod;
    for ( Int n=0 ; n<nbFrameAnalysis ; n++ ) {
      printf("analysis(fr-%d)\n", n*(1+nbSkipFrames) );
      pTVideoIOYuvInputFile0.read( pPicYuv0, aiPad0 );
      nbSkippedFrames++;

      if ( nbSkipFrames && (nbSkippedFrames+nbSkipFrames)<iNbFrames ) {
        pTVideoIOYuvInputFile0.skipFrames( nbSkipFrames, pPicYuv0->getWidth() - aiPad0[0], pPicYuv0->getHeight() - aiPad0[1] );
        nbSkippedFrames += nbSkipFrames;
      }

      toneMap( pPicYuv0, m_lutRef, sps );
    }  

#if SEI_COLOR_LUT
    pTVideoIOYuvInputFile0.goBack( nbSkippedFrames, pPicYuv0->getWidth() - aiPad0[0], pPicYuv0->getHeight() - aiPad0[1] );
#endif

    m_lutRef->prune( m_bVerbose );
  }

  //--------------------------------------------
  //  Prepare encoding
  //--------------------------------------------
  if ( m_model == MODEL_3DLUT ) {
    //m_lutRef->setDefault();

    printf("3D lut coding...\n");

    m_lutCoded->reset();
    m_lutResiduals->reset();

    Int errMax  = 0;
/*
    m_lutCoded->encode( errMax, m_errQ, m_lutRef, m_lutResiduals, false );

    m_lutCoded->reset();
    m_lutResiduals->reset();
*/
    m_lutCoded->encode( errMax, m_errQ, m_lutRef, m_lutResiduals, m_bVerbose );

    //writeFile("encoded_toneMapping");

    printf("3D lut coding done.\n");
  }
}

Void	TComToneMapping::encodeLUT()
{
  m_lutRef->reset( false, true, true );
  m_lutCoded->reset();
  m_lutResiduals->reset();

  Int errMax  = 0;
  m_lutCoded->encode( errMax, m_errQ, m_lutRef, m_lutResiduals, m_bVerbose );
}

Void	TComToneMapping::toneMap( TComPicYuv* pcBasePic, TComLut3D *lut, TComSPS* sps )
{
  printf("toneMap (%s model)...\n", "3D-LUT" );

  if ( !lut ) lut = m_lutCoded;

  if ( !m_pcTempPic ) {
	TComSPS* pcSps = NULL;
#if OBSOLETE_COLOR_GAMUT_SCALABILITY
	if ( !sps ) {     
	  pcSps = new TComSPS();
	  pcSps->setPicCropBottomOffset( pcBasePic->getPicCropBottomOffset() );
	  pcSps->setPicCropLeftOffset  ( pcBasePic->getPicCropLeftOffset()   );
	  pcSps->setPicCropRightOffset ( pcBasePic->getPicCropRightOffset()  );
	 /pcSps->setPicCropTopOffset   ( pcBasePic->getPicCropTopOffset()    );
	}
#endif
	m_pcTempPic = new TComPicYuv;
    m_pcTempPic->create( pcBasePic->getWidth(), pcBasePic->getHeight(), g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth, sps ? sps : pcSps );
    
    if ( pcSps )  delete pcSps;
  }
  else
    assert( pcBasePic->getWidth()==m_pcTempPic->getWidth() && pcBasePic->getHeight()==m_pcTempPic->getHeight() );


  Int iHeight   = pcBasePic->getHeight();
  Int iWidth    = pcBasePic->getWidth();
  Int iStride   = pcBasePic->getStride();
  Int iCStride  = pcBasePic->getCStride();

  Pel*	Lum0    = pcBasePic->getLumaAddr();
  Pel*	Cb0	    = pcBasePic->getCbAddr();
  Pel*	Cr0	    = pcBasePic->getCrAddr();
  Pel*	Lum1    = m_pcTempPic->getLumaAddr();
  Pel*	Cb1	    = m_pcTempPic->getCbAddr();
  Pel*	Cr1	    = m_pcTempPic->getCrAddr();

/*
	assert (sps);
  m_internalBitDepthY = sps->getBitDepthY();
  m_internalBitDepthC = sps->getBitDepthC(); 
*/
  m_internalBitDepthY = g_bitDepthY;
  m_internalBitDepthC = g_bitDepthC;

  Pel MaxY = (1 << m_internalBitDepthY) - 1;
  Pel MaxC = (1 << m_internalBitDepthC) - 1;
#if USE_IBD
  lut->set_ibd( m_internalBitDepthY, m_internalBitDepthC );
#endif

  node  nodeIn, nodeOut;
  for( Int y = 0; y < iHeight ; y++ ) {
    for( Int x = 0; x < iWidth ; x++ ) {
      nodeIn.sY = Lum0[x];
      nodeIn.sU = Cb0[x>>1];
      nodeIn.sV = Cr0[x>>1];

#if USE_IBD
      lut->interpol_ibd( &nodeIn, &nodeOut );
#else
      lut->interpol( &nodeIn, &nodeOut );
#endif

      Lum1[x]   = xClip( nodeOut.sY, MaxY );
      Cb1[x>>1] = xClip( nodeOut.sU, MaxC );
      Cr1[x>>1] = xClip( nodeOut.sV, MaxC );

      //if ( y==(iHeight-1) && x==0 ) printf("DEBUG\tTComToneMapping::toneMap():\t%d %d %d\n", Lum1[x], Cb1[x>>1], Cr1[x>>1] );
    }
    Lum0  += iStride;
    Lum1  += iStride;
    if ( (y % 2) ) {
      Cb0 += iCStride;
      Cr0 += iCStride;
      Cb1 += iCStride;
      Cr1 += iCStride;
    }
  }

  //dumpPicture("toneMapped", m_pcTempPic, 0 );

  printf("toneMap done.\n");
}

Void  TComToneMapping::dumpPicture(Char* fileName, TComPicYuv*	pPic, Int iShift )
{
  Char  name[1000];
  sprintf(name, "%s_%dx%d.yuv", fileName, pPic->getWidth(), pPic->getHeight());
  FILE	*fic = fopen(name, "wb");
  Pel *p = pPic->getLumaAddr();
  Int iStride = pPic->getStride();
  UChar*  uBuf = new UChar [pPic->getWidth()];
  for ( Int i=0 ; i<pPic->getHeight() ; i++ ) {
    for ( Int j=0 ; j<pPic->getWidth() ; j++ )
      uBuf[j] = (UChar)( p[j] >> iShift );
    fwrite( uBuf, sizeof(UChar), pPic->getWidth(), fic);
    p += iStride;
  }

  p = pPic->getCbAddr();
  iStride = pPic->getCStride();
  for ( Int i=0 ; i<pPic->getHeight()>>1 ; i++ ) {
    for ( Int j=0 ; j<pPic->getWidth()>>1 ; j++ )
      uBuf[j] = (UChar)( p[j] >> iShift );
    fwrite( uBuf, sizeof(UChar), pPic->getWidth()>>1, fic);
    p += iStride;
  }

  p = pPic->getCrAddr();
  iStride = pPic->getCStride();
  for ( Int i=0 ; i<pPic->getHeight()>>1 ; i++ ) {
    for ( Int j=0 ; j<pPic->getWidth()>>1 ; j++ )
      uBuf[j] = (UChar)( p[j] >> iShift );
    fwrite( uBuf, sizeof(UChar), pPic->getWidth()>>1, fic);
    p += iStride;
  }

  delete uBuf;

  fclose(fic);
}

inline  Int TComToneMapping::getNbp()
{
  return nbpCode[m_nbpCode];
}

Void	TComToneMapping::decode()
{
  m_lutCoded->reset();
  m_lutCoded->decode( 0, 0,0,0, m_errQ, m_lutResiduals, m_lutCoded->getOctTree(), m_bVerbose );
}
