/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComToneMapping.h
    \brief    Tone Mapping class (header)
*/

#ifndef __TCOMTONEMAPPING__
#define __TCOMTONEMAPPING__


// Include files
#include "../TLibVideoIO/TVideoIOYuv.h"

#include "TComPic.h"
#include "TComLut3D.h"

#define MODEL_3DLUT   (1)

// ====================================================================================================================
// Class definition
// ====================================================================================================================
/// tone mapping class
class TComToneMapping
{
  Bool        m_bVerbose;
  Char*       m_toneMappingFile;
  Int         m_model;            // MODEL_3DLUT
  Int         m_nbpCode;
  Int         m_internalBitDepthY;
  Int         m_internalBitDepthC;
	TComPicYuv*	m_pcTempPic;
  TComLut3D*  m_lutRef;
  TComLut3D*  m_lutCoded;
  TComLut3D*  m_lutResiduals;
  Int         m_errQ;

  Int         m_colour_map_id;
  Bool        m_colour_map_cancel_flag;
  Int         m_colour_map_repetition_period;
  Bool        m_colour_map_video_signal_type_present_flag;
  Int         m_colour_map_video_format;
  Bool        m_colour_map_video_full_range_flag;
  Bool        m_colour_map_description_present_flag;
  Int         m_colour_map_primaries;
  Int         m_colour_map_transfer_characteristics;
  Int         m_colour_map_matrix_coeffs;
  Int         m_bit_depth_lut_minus8;

  Void  loadFile( Char *cFile );
  Void  writeFile( Char *cFile );

  inline  Pel xClip( Pel x );
  inline  Pel xClip( Pel x, Pel Max );

public:
  TComToneMapping( Int iModel, Int iNbpCode, Char *cFile );
  ~TComToneMapping();

  Void        analysis( TVideoIOYuv& pTVideoIOYuvInputFile0, TComPicYuv* pPicYuv0, Int *aiPad0, TComSPS *sps=NULL, Int iPeriod=10, Int iNbAnalysis=20 );
  Void        encodeLUT();
  Void        decode();

  Void        dumpPicture( Char* fileName, TComPicYuv*	pPic, Int iShift );
  Void	      toneMap( TComPicYuv*, TComLut3D *lut=NULL, TComSPS* sps=NULL );

  inline Void setColour_map_id(Int val) { m_colour_map_id = val;};
  inline Void setColour_map_cancel_flag(Bool val) { m_colour_map_cancel_flag = val;};
  inline Void setColour_map_repetition_period(Int val) { m_colour_map_repetition_period = val;};
  inline Void setColour_map_video_signal_type_present_flag(Bool val) { m_colour_map_video_signal_type_present_flag = val;};
  inline Void setColour_map_video_format(Int val) { m_colour_map_video_format = val;};
  inline Void setColour_map_video_full_range_flag(Bool val) { m_colour_map_video_full_range_flag = val;};
  inline Void setColour_map_description_present_flag(Bool val) { m_colour_map_description_present_flag = val;};
  inline Void setColour_map_primaries(Int val) { m_colour_map_primaries = val;};
  inline Void setColour_map_transfer_characteristics(Int val) { m_colour_map_transfer_characteristics = val;};
  inline Void setColour_map_matrix_coeffs(Int val) { m_colour_map_matrix_coeffs = val;};
  inline Void setBit_depth_lut_minus8(Int val) {m_bit_depth_lut_minus8 = val;};

  inline Int getColour_map_id() { return m_colour_map_id;};
  inline Bool getColour_map_cancel_flag() { return m_colour_map_cancel_flag;};
  inline Int getColour_map_repetition_period() { return m_colour_map_repetition_period;};
  inline Bool getColour_map_video_signal_type_present_flag() { return m_colour_map_video_signal_type_present_flag;};
  inline Int getColour_map_video_format() { return m_colour_map_video_format;};
  inline Bool getColour_map_video_full_range_flag() { return m_colour_map_video_full_range_flag;};
  inline Bool getColour_map_description_present_flag() { return m_colour_map_description_present_flag;};
  inline Int getColour_map_primaries() { return m_colour_map_primaries;};
  inline Int getColour_map_transfer_characteristics() { return m_colour_map_transfer_characteristics;};
  inline Int getColour_map_matrix_coeffs() { return m_colour_map_matrix_coeffs;};
  inline Int getBit_depth_lut_minus8() { return m_bit_depth_lut_minus8;};

  inline Int  getNbpCode()      { return m_nbpCode;};
  inline Void setNbpCode(Int n) { m_nbpCode = n;};
  inline Int  getNbp();

  TComPicYuv* getToneMap()    { return m_pcTempPic; };

  inline Int        getModel()          { return m_model; };
  inline Void       setModel( Int im )  { m_model = im; };
  inline TComLut3D* getLutResiduals()   { return m_lutResiduals; };
  inline octTree*   getOctTree()        { return m_lutCoded->getOctTree(); };
  inline Int        getErrQ()           { return m_errQ; };
  inline Void       setErrQ( Int e )    { m_errQ = e; };
};

inline  Pel TComToneMapping::xClip( Pel x )
{ 
#define Max 255
  Pel pel = (Pel)( (x < 0) ? 0 : (x > Max) ? Max : x );
  return( pel );
}

inline  Pel TComToneMapping::xClip( Pel x, Pel max )
{ 
  Pel pel = (Pel)( (x < 0) ? 0 : (x > max) ? max : x );
  return( pel );
}

#endif 
